"""**Fallacy Removal** Chain runs a self-review of logical fallacies
as determined by paper
[Robust and Explainable Identification of Logical Fallacies in Natural
Language Arguments](https://arxiv.org/pdf/2212.07425.pdf).
It is modeled after `Constitutional AI` and in the same format, but applying logical
fallacies as generalized rules to remove them in output.
"""
